/*	--- Copyright University of Sussex 1990.  All rights reserved. ---------
 >	File:			C.all/lib/auto/newanysparse.p
 >	Purpose:		Create sparse arrays, using a tree of properties
 >	Author:			Jonathan Laventhol, 1984 (see revisions)
 >	Documentation:	REF *PROPS, HELP *NEWANYSPARSE (See below also)
 >	Related Files:	NEWSPARSE
 */
compile_mode:pop11 +strict;



/*
Makes a new sparse array, out of a tree of properties
the sparse array is a procedure (with updater) with N arguments if
it has N dimensions.  the pdprops of it are [sparse_array ^default]
(although note that changing the pdprops has no effect on the
behaviour of the array)
the tree is fleshed out the minimal amount to contain all the entries,
but it never contracts.	 you can control the size of the properties at
the different levels of the tree, and you can set the default for the
array.	you can also have "active defaults" -- where the value for an
unspecified entry is computed rather than stored.  the default size of the
properties is 20.  because the tree is made from properties, you can use
any pop item as a subscript, but remember that the subscript you access
with must be == with the subscript you stored with.	 (usually this means
strings, list, vectors are no good -- use words, numbers, procedures, undef
variables).

Some examples:

newanysparse(3, [])
	makes a three-dimensional array, default is []
newanysparse(2, undef)
	two dimensions, default is undef
newanysparse([10 20], 0)
	two dimensions, default is 0
	the list specifies that for the first dimension, there are going to be
	about 10 entries.  for the second dimension, about 20.
	(this controls the size of the properties used at each level)
newanysparse(5, active_default, apply)
	five dimensions
	if active_default is a procedure (at the time of this call) then
	the default for each element will be computed when needed, by applying
	active_default to the subscripts of the entry.
newanysparse(3, consvector(% 3 %), apply)
	three dimensions.  the default for each entry is {I J K} where I, J and
	K are the subscripts of that entry
newanysparse([10], apply);
	one dimension, default is the procedure apply
	this is the same in effect as newproperty([], 10, apply, true)
*/


section;


define global newanysparse(Ndims, default) -> pdr;
	lvars var prop Ndims prop fail success varslist PROP default i
		smart_default newval nextprop nextlab dim dimlist pdr PDPROPS;

	false -> smart_default;

	;;; want pdr applied at runtime for default?
	if default == apply and isprocedure(Ndims) then
		consstring(0) -> default;		;;; make unique marker
		Ndims -> smart_default;
		-> Ndims
	endif;

	;;; check dimension specifier is okay
	if isinteger(Ndims) then
		[] -> dimlist;
		repeat Ndims times 20 :: dimlist -> dimlist endrepeat;
	elseif islist(Ndims) then
		Ndims -> dimlist;
		length(Ndims) -> Ndims;
		for i in dimlist do
			unless isinteger(i) and i > 0 then
				mishap(i, 1, 'DIMENSION SIZE MUST BE POSITIVE INTEGER')
			endunless;
		endfor;
	else
		mishap(Ndims, 1, 'DIMENSION SPECIFIER MUST BE INTEGER OR LIST')
	endif;
	unless Ndims > 0 then
		mishap(Ndims, 1, 'NUMBER OF DIMENSIONS MUST BE 1 OR MORE');
	endunless;

	;;; check default procedure, if being used
/*	if smart_default and pdnargs(smart_default) /== Ndims then
		mishap(pdnargs(smart_default), 1,
					'DEFAULT PROCEDURE HAS WRONG NUMBER OF ARGUMENTS')
	endif; */

	;;; make top layer of array
	newproperty([], hd(dimlist), default, true) -> PROP;

	[sparse_array
	 %if smart_default then smart_default, apply else default endif
	%] -> PDPROPS;

	;;; build it
	sysPROCEDURE(PDPROPS, Ndims);
	sysNEW_LABEL() -> fail;		;;; where to go if default wanted
	sysNEW_LVAR() -> prop;		;;; this one first -- so it is in register
	[% repeat Ndims times sysNEW_LVAR() endrepeat %] -> varslist;
	applist(varslist, sysPOP);	;;; get the arguments
	ncrev(varslist) -> varslist;	;;; reverse for usage
	sysPUSHQ(PROP);				;;; get the top level property
	sysPOP(prop);
	for var in varslist do		;;; for each dimension
		sysPUSH(var);			;;;		prop(I) -> prop;
		sysCALL(prop);
		sysPOP(prop);

		sysPUSH(prop);			;;;		if prop == default then goto FAIL
		sysPUSHQ(default);
		sysCALL("==");
		sysIFSO(fail);
	endfor;
	if smart_default then		;;; active default, so
		sysPUSH(prop);			;;; fallen through -- return(prop)
		sysNEW_LABEL() -> success;
		sysGOTO(success);
		sysLABEL(fail);			;;; came because didn*t find result
		applist(varslist, sysPUSH); ;;; default_pdr(x, y, ...)
		sysCALLQ(smart_default);
		sysLABEL(success);
	else
		sysLABEL(fail);			;;; prop will be default, or value
		sysPUSH(prop);			;;; return(prop)
	endif;

	sysENDPROCEDURE() -> pdr;

;;; make the updater

	sysPROCEDURE(PDPROPS, Ndims+1); ;;; +1 is newval
	sysNEW_LVAR() -> nextprop;		;;; these first so in registers
	sysNEW_LVAR() -> prop;			;;; nextprop more access than prop
	[% repeat Ndims times sysNEW_LVAR() endrepeat %] -> varslist;
	sysNEW_LVAR() -> newval;
	applist(varslist, sysPOP);
	sysPOP(newval);					;;; the new value
	sysPUSHQ(PROP);					;;; get the top-level property
	sysPOP(prop);

	ncrev(varslist) -> varslist;	;;; use-order is reverse of pop-order

	tl(dimlist) -> dimlist;
	;;; for all dimensions except the last
	for dim in dimlist do
		hd(varslist) -> var;
		sysPUSH(var);				;;; prop(I) -> nextprop;
		sysCALL(prop);
		sysPOP(nextprop);
		sysPUSH(nextprop);			;;; if nextprop == default then
		sysPUSHQ(default);
		sysCALL("==");
		sysNEW_LABEL() -> nextlab;
		sysIFNOT(nextlab);

		sysPUSHQ(nil);
		sysPUSHQ(dim);
		sysPUSHQ(default);
		sysPUSHQ(true); ;;; perm prop
		sysCALL("newproperty");		;;;		newproperty(...) -> nextprop;

		sysPOP(nextprop);			;;;		nextprop -> prop(I);
		sysPUSH(nextprop);
		sysPUSH(var);
		sysUCALL(prop);
	sysLABEL(nextlab);				;;; endif
		sysPUSH(nextprop);			;;; nextprop -> prop;
		sysPOP(prop);
	tl(varslist) -> varslist;
	endfor;
	sysPUSH(newval);				;;; newval -> prop(z);
	sysPUSH(hd(varslist));			;;; (one left in this list)
	sysUCALL(prop);
	sysENDPROCEDURE() -> updater(pdr);
enddefine;


endsection;


/* --- Revision History ---------------------------------------------------
--- John Gibson, Aug 27 1990
		Changed n*ote to weak
--- John Gibson, Jun 30 1989
		Added +strict etc
--- John Gibson, Aug  5 1987
		sectioned
 */
