/* --- Copyright University of Sussex 1992. All rights reserved. ----------
 > File:            C.all/lib/lib/slowprocs.p
 > Purpose:         completely remove the fast procedures, for debugging.
 > Author:          Jonathan Laventhol, Jun 17 1983 (see revisions)
 > Documentation:   REF *FASTPROCS, HELP *EFFICIENCY
 */

#_TERMIN_IF DEF POPC_COMPILING

section;

;;; NB This file must be loaded BEFORE user files which use the fast procedures.
;;; NB Library fast_procedures must be loaded if library is to work is to work



;;; And words that contain these strings are assumed to be fast, and the
;;; corresponding word without the string is assumed to be the slow version
;;; (if the valof both are procedures)
lconstant fast_indicators = {'fi_' 'fast_'};

;;; These have to be left alone because of semantic differences between the
;;; fast and slow versions
lconstant exclusions = [% "'fast_bagof/3'", "'fast_setof/3'" %];


;;; Return the word with the sub-word deleted from it
define lconstant delete_from_word(sub_word, word);
	lvars sub_word, word, new_word;
	lvars sub_word_pos, end_sub_word_pos, word_len;

	issubstring(sub_word, word) -> sub_word_pos;
	sub_word_pos + length(sub_word) - 1 -> end_sub_word_pos;
	length(word) -> word_len;

	consword(#|
		explode(subword(1, sub_word_pos-1, word));
		explode(subword(end_sub_word_pos+1, word_len-end_sub_word_pos, word));
	|#);

enddefine;



;;; Returns all the fast and slow procedures that are defined in the form
;;; [<name_of_fast_proc> {<fast_proc> <slow_proc>} ]
define lconstant fast_and_slow(indicators);
	lvars indicators;
	lvars indicator, num_of_fast_items, fast_item, slow_item, valof_fast_item,
		valof_slow_item;

	for indicator in_vector indicators do;

		for fast_item in_list match_wordswith('@a' >< indicator) do;

			nextif(lmember(fast_item, exclusions));

			delete_from_word(indicator, fast_item) -> slow_item;

			if isdefined(fast_item) and isdefined(slow_item)
			then

				valof(fast_item) -> valof_fast_item;
				valof(slow_item) -> valof_slow_item;

				if isprocedure(valof_fast_item)
				and isprocedure(valof_slow_item)
				then
					[%fast_item, {%valof_fast_item, valof_slow_item%} %];
				endif;

			endif;

		endfor;

	endfor;

enddefine;



lconstant Slowprocs_table = newanyproperty(
	[% fast_and_slow(fast_indicators) %],
	100, 1, 75,
	false, false, true, false, false
);



define global constant procedure slowprocs(flag);
	lvars table;
	dlvars   flag;
	if flag then 2 else 1 endif -> flag;
	appproperty(Slowprocs_table,
				procedure(id, pdrs);
					lvars id pdrs;
					pdrs(flag) -> valof(id)
				endprocedure);
	if flag == 1 then
		'Fast procedures are fast'
	else
		'Fast procedures removed'
	endif =>
enddefine;

;;; Initialise - unprotect all words in the table, and turn slowprocs on

appproperty(Slowprocs_table, erase <> sysunprotect);

slowprocs(true);

endsection;

[] -> proglist;


/*  --- Revision History ---------------------------------------------------
--- Robert John Duncan, May 15 1992
		Changed to exclude fast Prolog predicates which are semantically
		distinct from their slow versions (see BR isl-fr.4431)
--- Adrian Howard, Jun 25 1991
		Altered so fast/checking procedures are now automatically found by
		looking for procedures with fi/fast prefix
--- John Gibson, Jun 30 1989
		Added all the missing things
--- Simon Nichols, Nov  7 1988
		Added fast prolog procedures and increased property table size from
		31 to 63.
--- John Gibson, Apr  1 1988
		Added -fast_frozval-
--- John Gibson, Mar 20 1988
		Added -fast_repeat-
--- John Gibson, Mar 16 1988
		Added -fast_sysread- and -fast_syswrite-
--- John Gibson, Mar 14 1988
		Added -fi_div- and -fi_rem- (now in system)
--- Aaron Sloman, May 25 1987
	changed "ident" to "id" as it would no longer compile
--- Aaron Sloman, Jun  3 1986 restored fast_for, after redefining endfast_for
	as a macro equivalent to endfor
--- John Williams, May 31 1986
		Commented out entries for 'fast_for' and 'endfast_for', since
		using 'for' instead won't accept 'endfast_for' as a terminator
--- Aaron Sloman, May 19 1986
		Previously referred to non-existent HELP * SLOWPROCS
--- Aaron Sloman, May 15 1986
		Put in additional entries and sorted alphabetically.
--- John Williams, Feb 25 1986
		Added entries for 'fi_' bitwise operators, FI_MIN, and FI_MAX.
		Also made slow/fast procedure table an lconstant
--- Mark  Rubinstein, Jun 25 1985
		Mapping for FAST_APPPROPERTY added.
--- Jonathan Laventhol, July 23 1985
		Extended  so that you can switch from the fast to the slow procedures;
 */
