#| --- Copyright University of Sussex 1995. All rights reserved. ----------
 | File:            C.all/lisp/src/lists.lsp
 | Purpose:         Common Lisp list functions
 | Author:          John Williams & Sam Valentine, Feb 26 1986 (see revisions)
 | Documentation:   CLtL, p262-281
 | Related Files:   C.all/lisp/src/lists.p
 |#

;;; Sorting out :TEST and :TEST-NOT arguments to list & sequence functions

(defun SYS:CHOOSE-SEQ-TEST1 (target test test-not)
	(if test
		(if test-not
			(error "Both :TEST and :TEST-NOT options supplied" test test-not)
			#'(lambda (seq-element) (funcall test target seq-element)))
		(if test-not
			(complement #'(lambda (seq-element)
							(funcall test-not target seq-element)))
			#'(lambda (seq-element) (eql target seq-element)))))


(defun SYS:CHOOSE-SEQ-TEST2 (test test-not)
	(if test
		(if test-not
			(error "Both :TEST and :TEST-NOT options supplied" test test-not)
			test)
		(if test-not
			(complement test-not)
			#'eql)))


;;; List functions

(defun TREE-EQUAL (x y &key test test-not)
	(setq test (sys:choose-seq-test2 test test-not))
	(labels
		((do-tree-equal (x y)
			(if (consp x)
				(and (consp y)
					 (do-tree-equal (car x) (car y))
					 (do-tree-equal (cdr x) (cdr y)))
				(and (atom y) (funcall test x y) t))))
		(do-tree-equal x y)))


(defun LIST-LENGTH (list)
	(do ((n 0 (+ n 2))
		 (fast list (cddr fast))
		 (slow list (cdr slow)))
		(nil)
		(when (endp fast) (return n))
		(when (endp (cdr fast)) (return (1+ n)))
		(when (and (eq fast slow) (plusp n)) (return nil))))


(defun MAKE-LIST (size &key initial-element &aux list)
	(dotimes (i size list)
		(setq list (cons initial-element list))))


(defun REVAPPEND (x y)
	(dolist (i x y) (setq y (cons i y))))


(defun NRECONC (x y &aux temp)
	(loop
		(when (endp x) (return y))
		;; use rotatef ??
		(setf temp (cdr x) (cdr x) y y x x temp)))


(defun LDIFF (list sublist)
	(do ((list list  (cdr list))
		 (result nil (cons (car list) result)))
		((or (endp list) (eq sublist list)) (sys:nreverse-list result))))


(defun SUBST-IF (new pred tree &key (key #'identity))
	(labels
		((do-subst (tree)
			(cond
				((funcall pred (funcall key tree)) new)
				((consp tree) (cons (do-subst (car tree))
									(do-subst (cdr tree))))
				(t tree))))
		(do-subst tree)))


(defun SUBST (new old tree &key test test-not (key #'identity))
	(subst-if new (sys:choose-seq-test1 old test test-not) tree :key key))


(defun SUBST-IF-NOT (new pred tree &key (key #'identity))
	(subst-if new (complement pred) tree :key key))


(defun NSUBST-IF (new pred tree &key (key #'identity))
	(labels
		((do-nsubst (tree)
			(cond
				((funcall pred (funcall key tree)) new)
				((consp tree)
					(setf (car tree) (do-nsubst (car tree)))
					(setf (cdr tree) (do-nsubst (cdr tree)))
					tree)
				(t tree))))
		(do-nsubst tree)))


(defun NSUBST-IF-NOT (new pred tree &key (key #'identity))
	(nsubst-if new (complement pred) tree :key key))


(defun NSUBST (new old tree &key test test-not (key #'identity))
	(nsubst-if new (sys:choose-seq-test1 old test test-not) tree :key key))


(defun SUBLIS (alist tree &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(labels
		((do-sublis (tree &aux pair)
			(cond
				((setq pair (assoc tree alist :test test :key key))
					(cdr pair))
				((consp tree)
					(cons (do-sublis (car tree)) (do-sublis (cdr tree))))
				(t tree))))
	(do-sublis tree)))


(defun NSUBLIS (alist tree &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(labels
		((do-nsublis (tree &aux pair)
			(cond
				((setq pair (assoc tree alist :test test :key key))
					(cdr pair))
				((consp tree)
					(setf (car tree) (do-nsublis (car tree)))
					(setf (cdr tree) (do-nsublis (cdr tree)))
					tree)
				(t tree))))
		(do-nsublis tree)))


(defun MEMBER (item list &key (test #'eql) test-not (key #'identity))
	(if test-not
		(loop
			(when (endp list) (return nil))
			(if (funcall test-not item (funcall key (car list)))
				(setq list (cdr list))
				(return list)))
		(loop
			(when (endp list) (return nil))
			(if (funcall test item (funcall key (car list)))
				(return list)
				(setq list (cdr list))))))


(defun MEMBER-IF (pred list &key (key #'identity))
	(loop
		(when (endp list) (return nil))
		(if (funcall pred (funcall key (car list))) (return list))
		(setq list (cdr list))))


(defun MEMBER-IF-NOT (pred list &key (key #'identity))
	(loop
		(when (endp list) (return nil))
		(unless (funcall pred (funcall key (car list))) (return list))
		(setq list (cdr list))))


(defun TAILP (sublist list)
	(loop
		(if (eql sublist list)
			(return t)
			(if (atom list)
				(return nil)
				(setq list (cdr list))))))


(defun ADJOIN (item list &rest keys &key (key #'identity) &allow-other-keys)
	(if (apply #'member (funcall key item) list keys)
		list
		(cons item list)))


(defun UNION (l1 l2 &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(dolist (i l1 l2)
		(unless (member (funcall key i) l2 :test test :key key)
				(setq l2 (cons i l2)))))


(defun NUNION (l1 l2 &key test test-not (key #'identity) &aux temp)
	(setq test (sys:choose-seq-test2 test test-not))
	(do ()
		((endp l1) l2)
		(if (member (funcall key (car l1)) l2 :test test :key key)
			(setq l1 (cdr l1))
			;; (rotatef l1 (cdr l1) l2)
			(setf temp (cdr l1) (cdr l1) l2 l2 l1 l1 temp))))


(defun INTERSECTION (l1 l2 &key test test-not (key #'identity) &aux result)
	(setq test (sys:choose-seq-test2 test test-not))
	(dolist (i l1 result)
		(if (member (funcall key i) l2 :test test :key key)
			(setq result (cons i result)))))


(defun NINTERSECTION (l1 l2 &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(do* ((rem (cons nil l1))
		  (result rem))
		 ((endp (cdr rem)) (cdr result))
		(if (member (funcall key (cadr rem)) l2 :test test :key key)
			(setq rem (cdr rem))
			(setf (cdr rem) (cddr rem)))))


(defun SET-DIFFERENCE (l1 l2 &key test test-not (key #'identity) &aux result)
	(setq test (sys:choose-seq-test2 test test-not))
	(dolist (i l1 result)
		(unless (member (funcall key i) l2 :test test :key key)
				(setq result (cons i result)))))


(defun NSET-DIFFERENCE (l1 l2 &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(do* ((rem (cons nil l1))
			(result rem))
		 ((endp (cdr rem)) (cdr result))
		(if (member (funcall key (cadr rem)) l2 :test test :key key)
			(setf (cdr rem) (cddr rem))
			(setq rem (cdr rem)))))


(defun SET-EXCLUSIVE-OR (l1 l2 &key test test-not (key #'identity)
							   &aux result)
	(setq test (sys:choose-seq-test2 test test-not))
	(dolist (i l1)
		(unless (member (funcall key i) l2 :test test :key key)
				(setq result (cons i result))))
	(dolist (i l2 result)
		(unless (member (funcall key i) l1 :test test :key key)
				(setq result (cons i result)))))


(defun NSET-EXCLUSIVE-OR (l1 l2 &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(do* ((list1 (cons nil l1))
		  (list2 (cons nil l2))
		  (p1 list1)
		  (p2 list2))
		 ((endp (cdr p1)) (setf (cdr p1) (cdr list2)) (cdr list1))
		(if (endp (cdr p2))
			(setq p1 (cdr p1) p2 list2)
			(if (funcall test (funcall key (cadr p1))
							  (funcall key (cadr p2)))
				(setf (cdr p1) (cddr p1)
					  (cdr p2) (cddr p2) p2 list2)
				(setq p2 (cdr p2))))))



(defun SUBSETP (l1 l2 &key test test-not (key #'identity))
	(setq test (sys:choose-seq-test2 test test-not))
	(dolist (i l1 t)
		(unless (member (funcall key i) l2 :test test :key key)
				(return nil))))


(defun ACONS (x y a-list)
	(cons (cons x y) a-list))


(defun PAIRLIS (keys data &optional a-list)
	(dolist (key keys a-list)
		(when (endp data) (return a-list))
		(setq a-list (acons key (car data) a-list) data (cdr data))))


(defun ASSOC (item a-list &key (test #'eql) test-not (key #'identity))
	(if test-not
		(dolist (pair a-list)
				(and (consp pair)
					 (not (funcall test-not item (funcall key (car pair))))
					 (return pair)))
		(dolist (pair a-list)
				(and (consp pair)
					 (funcall test item (funcall key (car pair)))
					 (return pair)))))


(defun ASSOC-IF (pred a-list &key (key #'identity))
	(dolist (pair a-list)
		(and (consp pair)
			 (funcall pred (funcall key (car pair)))
			 (return pair))))


(defun ASSOC-IF-NOT (pred a-list &key (key #'identity))
	(dolist (pair a-list)
		(and (consp pair)
			 (not (funcall pred (funcall key (car pair))))
			 (return pair))))


(defun RASSOC (item a-list &key (test #'eql) test-not (key #'identity))
	(if test-not
		(dolist (pair a-list)
				(and (consp pair)
					 (not (funcall test-not item (funcall key (cdr pair))))
					 (return pair)))
		(dolist (pair a-list)
				(and (consp pair)
					 (funcall test item (funcall key (cdr pair)))
					 (return pair)))))


(defun RASSOC-IF (pred a-list &key (key #'identity))
	(dolist (pair a-list)
		(and (consp pair)
			 (funcall pred (funcall key (cdr pair)))
			 (return pair))))


(defun RASSOC-IF-NOT (pred a-list &key (key #'identity))
	(dolist (pair a-list)
		(and (consp pair)
			 (not (funcall pred (funcall key (cdr pair))))
			 (return pair))))



#| --- Revision History ---------------------------------------------------
--- John Williams, May  5 1995
		SYS:CHOOSE-SEQ-TEST1 & SYS:CHOOSE-SEQ-TEST2 moved here from _______seq.lsp
--- John Williams, May 16 1994
		SUBST-IF-NOT and NSUBST-IF-NOT now use COMPLEMENT.
--- John Williams, Aug 27 1993
		Upgraded to Steele 1990.
 |#
