/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:            C.all/lisp/src/print.p
 > Purpose:         Printing procedures for built in data-types
 > Author:          John Williams, Nov 28 1985 (see revisions)
 > Documentation:   CLtL, p283-289
 > Related Files:   C.all/lisp/src/pprint.p, C.all/lisp/src/defstruct.p
 */

lisp_compile_mode;

section $-lisp;


constant procedure (isboundtoken, lisp_class_print, pr_sym_name);

vars procedure process_pid = not;       ;;; See C.all/lisp/modules/process.lsp


global vars
	;;; global 'cos used in FORMAT_PRINT
	depth_printed       =   0,
	length_printed      =   0,
	print_array         =   true,
	print_circle        =   nil,
	print_escape        =   true,
	print_gensym        =   true,
	print_miser_width   =   nil,            ;;; not used yet
	print_length        =   nil,
	print_level         =   nil,
	print_lines         =   nil,            ;;; not used yet
	print_pretty        =   true,
	print_radix         =   nil,
	print_readably      =   nil,
	print_right_margin  =   nil,
	print_pprint_dispatch = nil,            ;;; not used yet
	;


global vars active (
	print_base = integer_range_variable
					(% ident pop_pr_radix, 2, 36, @*PRINT-BASE* %),
	);


define print_not_readable(item);
	lvars item;
	dlocal print_readably = nil, read_eval = true;
	lisp_error('Print not readable', [^item])
enddefine;


define macro CHECK_PRINT_READABLY(item);
	lvars item;
	[if print_readably /== nil then
		print_not_readable(^item)
	 endif;].dl
enddefine;


define macro CHECK_PR_LEVEL;
	[if isinteger(print_level)
	 and depth_printed fi_>= print_level
	 and print_readably == nil
	 then
		cucharout(`#`);
		return
	 else
		depth_printed fi_+ 1 -> depth_printed
	 endif;].dl
enddefine;


define macro CHECK_PR_LEN;
	[if isinteger(print_length)
	 and length_printed fi_>= print_length
	 and print_readably == nil
	 then
		cucharout(`.`);
		cucharout(`.`);
		cucharout(`.`);
		cucharout(`)`);
		return
	 else
		length_printed fi_+ 1 -> length_printed
	 endif;].dl
enddefine;


/* Core print procedures */

define lconstant Pr_pop11_simple(item);
	lvars item;
	CHECK_PRINT_READABLY item;
	cucharout(`#`);
	sys_syspr(item)
enddefine;


define lconstant Print_by_key(item);
	lvars item;
	lisp_class_print(datakey(item))(item)
enddefine;


define default_print_object(item, stream) -> item;
	lvars item, stream;
	dlocal cucharout, standard_output;
	unless stream == standard_output do
		stream -> standard_output;
		recursive_str_output(stream) -> cucharout
	endunless;
	Print_by_key(item)
enddefine;


define lconstant Try_print_object(item);
	lvars item, p;
	sf_token(@PRINT-OBJECT) -> p;
	if isboundtoken(p)
	and generic_function_p(ft_valof(p) ->> p) then
		lisp_apply(item, standard_output, p, 2, 0);
		true
	else
		false
	endif
enddefine;


define lispsyspr(item);
	lvars item;
	if is_poplog_item(item) then
		unless Try_print_object(item) do
			Print_by_key(item)
		endunless
	else
		Pr_pop11_simple(item)
	endif
enddefine;


vars procedure _lisppr = lispsyspr;


/* Utilities */

lconstant procedure Popintpr = sys_syspr;


define lconstant Popfloatpr() with_nargs 1;
	dlocal pop_pr_radix = 10, pop_pr_exponent = false;
	sys_syspr()
enddefine;


define lconstant Decpr() with_nargs 1;
	dlocal pop_pr_radix = 10;
	Popintpr()
enddefine;


lconstant Pr_unreadable_quote = '*PR-UNREADABLE-QUOTE*';

define lconstant Pr_unreadable();
	lvars list, item;
	sysconslist() -> list;
	cucharout(`#`);
	cucharout(`<`);
	repeat
		sys_grbg_destpair(list) -> (item, list);
		if item == Pr_unreadable_quote and ispair(list) then
			_lisppr(sys_grbg_destpair(list) -> list)
		elseif isstring(item) then
			procedure() with_nargs 1;
				dlocal print_escape = nil;
				pr_sym_name()
			endprocedure(item)
		else
			_lisppr(item)
		endif;
		quitif(list == []);
		cucharout(`\s`)
	endrepeat;
	cucharout(`>`)
enddefine;


/* Printing arrays, vectors, strings etc */

define lconstant Short_array_pr(a);
	lvars a;
	Pr_unreadable(popstackmark, 'ARRAY',
					array_element_type(a), array_dimensions(a))
enddefine;


define lconstant Pr_bv_between(bv, lo, hi);
	lvars bv lo hi i;
	cucharout(`#`);
	cucharout(`*`);
	fast_for i from lo to hi do
		cucharout(fast_subscrbitvector(i, bv) fi_+ `0`)
	endfast_for
enddefine;


define lconstant Pr_bitvector(bv);
	lvars bv;
	if print_array /== nil
	or print_readably /== nil then
		Pr_bv_between(bv, 1, fast_vector_length(bv))
	else
		Short_array_pr(bv)
	endif
enddefine;


define lconstant Ppr_string_between(str, lo, hi);
	lvars c hi i lo str;
	cucharout(`"`);
	fast_for i from lo to hi do
		fast_subscrs(i, str) -> c;
		if c == `\\` or c == `"` then
			cucharout(`\\`)
		endif;
		cucharout(c)
	endfast_for;
	cucharout(`"`)
enddefine;


define lconstant Pr_string_between(str, lo, hi);
	lvars hi i lo str;
	if print_escape /== nil
	or print_readably /== nil then
		Ppr_string_between(str, lo, hi)
	else
		fast_for i from lo to hi do
			cucharout(fast_subscrs(i, str))
		endfast_for
	endif
enddefine;


define lconstant Pr_string(str);
	lvars str;
	Pr_string_between(str, 1, fast_vector_length(str))
enddefine;


define lconstant Pr_vec_between(vec, lo, hi);
	lvars hi i lo procedure subscr vec;
	dlocal length_printed = 0;

	class_fast_subscr(datakey(vec)) -> subscr;
	cucharout(`(`);
	fast_for i from lo to hi do
		CHECK_PR_LEN;
		_lisppr(subscr(i, vec));
		unless i == hi do
			cucharout(`\s`)
		endunless
	endfast_for;
	cucharout(`)`)
enddefine;


define lconstant Pr_vector(vec);
	lvars vec;
	dlocal depth_printed;

	if print_array /== nil
	or print_readably /== nil then
		unless isvector(vec) do
			CHECK_PRINT_READABLY vec
		endunless;
		CHECK_PR_LEVEL;
		cucharout(`#`);
		Pr_vec_between(vec, 1, fast_vector_length(vec))
	else
		Short_array_pr(vec)
	endif
enddefine;


define lconstant Pr_array(a);
	lvars a rank;
	dlvars I procedure Printer Vec;

	if print_array == nil
	and print_readably == nil
	then
		chain(a, Short_array_pr)
	endif;

	array_rank(a) -> rank;
	if rank == 0 then
		cucharout(`#`);
		Decpr(0);
		cucharout(`A`);
		cucharout(`\s`);
		_lisppr(a());
		return
	endif;

	arrayvector(a) -> Vec;
	if isvector(Vec) then
		Pr_vec_between
	elseif isstring(Vec) then
		if rank == 1 then
			Pr_string_between
		else
			Ppr_string_between
		endif
	elseif isbitvector(Vec) then
		Pr_bv_between
	else
		CHECK_PRINT_READABLY a;
		Pr_vec_between
	endif -> Printer;

	if rank == 1 then
		if Printer == Pr_vec_between then
			cucharout(`#`)
		endif;
		Printer(Vec, arrayseq_bounds(a))
	else
		define lconstant Array_pr(bounds);
			lvars bounds hi lo;
			dlocal depth_printed, length_printed = 0;

			CHECK_PR_LEVEL;
			fast_destpair(fast_destpair(bounds)) -> bounds -> hi -> lo;
			if bounds == [] then
				I fi_+ (hi fi_- lo) -> hi;
				Printer(Vec, I, hi);
				hi fi_+ 1 -> I
			else
				cucharout(`(`);
				fast_for lo from 0 to (hi fi_- lo) do
					CHECK_PR_LEN;
					Array_pr(bounds);
					unless lo == hi do
						cucharout(`\s`)
					endunless
				endfast_for;
				cucharout(`)`)
			endif
		enddefine;

		arrayvector_bounds(a) -> I ->;
		cucharout(`#`);
		Decpr(rank);
		cucharout(`A`);
		Array_pr(boundslist(a))
	endif
enddefine;


/* Printing Lists */

define lconstant Pr_list(list);
	lvars list save;
	dlocal depth_printed, length_printed = 0;

	CHECK_PR_LEVEL;
	if print_pretty /== nil then
		fast_front(list) -> save;
		if (save == @QUOTE or save == @FUNCTION)
		and islistlength(list, 2) then
			if save == @FUNCTION then cucharout(`#`) endif;
			cucharout(`'`);
			chain(fast_front(fast_back(list)), _lisppr)
		endif
	endif;
	list -> save;
	cucharout(`(`);
	repeat
		CHECK_PR_LEN;
		_lisppr(fast_destpair(list) -> list);
		quitif(atom(list) or list == save);
		cucharout(`\s`)
	endrepeat;
	if list /== [] then
		cucharout(`\s`);
		cucharout(`.`);
		cucharout(`\s`);
		_lisppr(list)
	endif;
	cucharout(`)`)
enddefine;


/* Printing symbol names */

define lconstant No_escapes_needed(s, isalpha);
	lvars s, procedure isalpha, l, c, i;
	fast_vector_length(s) -> l;
	returnif(l == 0) (false);
	subscrs(1, s) -> c;
	if c == `+` or c == `-` then
		returnif(l == 1) (true);
		subscrs(2, s) -> c
	endif;
	if isdigitcode(c, pop_pr_radix) then
		subscrs(l, s) -> c;
		if isdigitcode(c, pop_pr_radix) or c == `.` then
			return(false)
		endif
	endif;
	for c with_index i in_string s do
		unless isalpha(c)
				or (c fi_>= `0` and c fi_<= `9`)
				or strmember(c, '&*+-/<=>_')
		do
			return(false)
		endunless
	endfor;
	true
enddefine;


define lconstant Escape_pr(s);
	lvars s;
	cucharout(`|`);
	appdata(s, cucharout);
	cucharout(`|`);
enddefine;


lvars
	First               =   true,
	Print_case          =   @:UPCASE,
	procedure
	Print_case_convert  =   lowertoupper,
	;


define global active print_case;
	Print_case
enddefine;


define updaterof active print_case(new);
	lvars new;
	if new == @:UPCASE then
		lowertoupper
	elseif new == @:DOWNCASE then
		uppertolower
	elseif new == @:CAPITALIZE then
		procedure(c);
			lvars c;
			if First then lowertoupper(c) else uppertolower(c) endif;
			not(isalphacode(c) or isnumbercode(c)) -> First
		endprocedure
	elseif new == nil then
		identfn
	else
		warn('Ignoring illegal value for *PRINT-CASE*', [^new]) ->;
		return
	endif -> Print_case_convert;
	new -> Print_case
enddefine;


define lconstant Pr_sym_name(s);
	lvars s, c;
	dlocal First = true;
	for c in_string s do
		cucharout(Print_case_convert(c))
	endfor
enddefine;


define pr_sym_name(name);
	lvars name, rcase, isalpha, c;
	dlocal Print_case_convert;

	if print_escape == nil then
		Pr_sym_name(name)
	else
		readtable_case_sym(readtable) -> rcase;
		if rcase == @:UPCASE then
			isuppercode
		elseif rcase == @:DOWNCASE then
			islowercode
		else
			isalphacode
		endif -> isalpha;
		if No_escapes_needed(name, isalpha) then
			if isalpha == isalphacode then
				if rcase == @:INVERT
				and (all_same_case(name) ->> c)
				and Print_case_convert /== identfn
				then
					if isuppercode(c) then
						uppertolower
					else
						lowertoupper
					endif -> Print_case_convert;
					Pr_sym_name(name)
				else
					appdata(name, cucharout)
				endif
			else
				Pr_sym_name(name)
			endif
		else
			Escape_pr(name)
		endif
	endif
enddefine;


define lconstant Pr_package(pkg);
	lvars pkg, name;
	CHECK_PRINT_READABLY pkg;
	package_name(pkg) -> name;
	Pr_unreadable(
		popstackmark,
		unless name == nil then
			name
		endunless,
		'PACKAGE',
		if name == nil then
			'@', address_of(package)
		endif)
enddefine;


define lconstant Pr_symbol(sym);
	lvars sym, name, pkg, pname, status;
	symbol_string(sym) -> name;
	symbol_package(sym) -> pkg;
	if pkg == nil then
		if print_readably /== nil
		or (print_escape /== nil and print_gensym /== nil)
		then
			cucharout(`#`);
			cucharout(`:`)
		endif
	elseif print_escape /== nil
	or print_readably /== nil
	then
		if pkg == keyword_package then
			cucharout(`:`)
		else
			unless pkg == package
			or (find_symbol(name, package) ->) == sym do
				if (package_name(pkg) ->> pname) == nil then
					Pr_package(pkg)
				else
					pr_sym_name(pname)
				endif;
				cucharout(`:`);
				find_symbol(name, pkg) -> (, status);
				if status == @:INTERNAL then
					cucharout(`:`)
				endif
			endunless
		endif
	endif;
	pr_sym_name(name)
enddefine;


procedure(sym);
	lvars sym;
	dlocal print_case, print_escape = true, print_gensym = true;
	@:UPCASE -> print_case;
	cucharout(`@`);
	Pr_symbol(sym);
endprocedure -> class_print(symbol_key);



/* Printing numbers */

define lconstant Pr_float(num);
	lvars expchar expt num;
	dlocal
		popdprecision = true,
		pop_pr_places,
		;

	if issimple(num) then
		#_<intof(float_digits(decimal_0) * log(2) / log(10))>_#,
		(read_default_float_format == @DOUBLE-FLOAT
		 or read_default_float_format == @LONG-FLOAT) and `f`
	else
		#_<intof(float_digits(ddecimal_0) * log(2) / log(10))>_#,
		(read_default_float_format == @SINGLE-FLOAT
		 or read_default_float_format == @SHORT-FLOAT) and `d`
	endif -> expchar -> pop_pr_places;

	if num < 0 then
		negate(num) -> num;
		cucharout(`-`)
	endif;

	0 -> expt;
	unless sys_=(num, 0) do
		if num < 1e-3 or num >= 1e7 then
			intof(log10(num)) -> expt;
			num / (10 ** expt) -> num
		endif
	endunless;

	Popfloatpr(num);
	if expchar then
		cucharout(expchar);
		Decpr(expt)
	elseunless expt == 0 do
		cucharout(`e`);
		Decpr(expt)
	endif
enddefine;


define lconstant Pr_radix();
	cucharout(`#`);
	if pop_pr_radix == 2 then
		`b`
	elseif pop_pr_radix == 8 then
		`o`
	elseif pop_pr_radix == 16 then
		`x`
	else
		Decpr(pop_pr_radix);
		`r`
	endif;
	cucharout()
enddefine;


define lconstant Pr_integer(int);
	lvars int;
	if print_radix == nil then
		Popintpr(int)
	elseif pop_pr_radix == 10 then
		Popintpr(int);
		cucharout(`.`)
	else
		Pr_radix();
		Popintpr(int)
	endif
enddefine;


define lconstant Pr_ratio(rat);
	lvars rat;
	if print_radix /== nil then
		Pr_radix()
	endif;
	Popintpr(destratio(rat) -> rat);
	cucharout(`/`);
	Popintpr(rat);
enddefine;


define lconstant Pr_complex(cx);
	lvars cx;
	cucharout(`#`);
	cucharout(`C`);
	cucharout(`(`);
	_lisppr(destcomplex(cx) -> cx);
	cucharout(`\s`);
	_lisppr(cx);
	cucharout(`)`)
enddefine;


/* Unreadable objects, booleans, etc */

define lconstant Pr_bytespec(spec);
	lvars spec;
	lconstant list = writeable [^@BYTE 0 0];
	if read_eval == nil then
		CHECK_PRINT_READABLY spec
	endif;
	fast_frozval(1, spec) -> fast_front(fast_back(list));
	fast_frozval(2, spec) -> fast_front(fast_back(fast_back(list)));
	cucharout(`#`);
	cucharout(`.`);
	_lisppr(list)
enddefine;


define lconstant Pr_pathname(path);
	lvars path, s;
	namestring(path) -> s;
	if print_escape == nil
	and print_readably == nil then
		appdata(s, cucharout)
	else
		cucharout(`#`);
		cucharout(if pt_device(path) == @:LOGICAL then `L` else `P` endif);
		_lisppr(s)
	endif
enddefine;


define lconstant Pr_random_state(rs);
	lvars rs seed;
	lconstant list = writeable [^@RANDOM-STATE 0];
	random_seed(rs) -> seed;
	unless seed do
		apply_random_state(pop_max_int, rs) ->;
		random_seed(rs) -> seed
	endunless;
	seed -> fast_front(fast_back(list));
	cucharout(`#`);
	cucharout(`S`);
	_lisppr(list)
enddefine;


define lconstant Pr_boolean() with_nargs 1;
	if () then
		Pr_symbol(true)
	elseif print_readably /== nil then
		Pr_symbol(@POP11:FALSE)
	else
		Pr_unreadable(popstackmark, 'FALSE')
	endif
enddefine;


define lconstant Pr_character(char);
	lvars char name;
	if print_escape /== nil
	or print_readably /== nil
	then
		cucharout(`#`);
		cucharout(`\\`);
		if (char_name(char) ->> name) /== nil then
			appdata(name, cucharout);
			return
		endif;
	endif;
	cucharout(fast_char_code(char))
enddefine;


define lconstant Pr_hash_table(h);
	lvars h;
	CHECK_PRINT_READABLY h;
	Pr_unreadable(
		popstackmark,
		'HASHTABLE',
		if issymbol(pdprops(h)) then
			@:TEST,
			pdprops(h)
		endif,
		@:SIZE,
		property_size(h))
enddefine;


define lconstant Pr_nil() with_nargs 1;
	->;
	if print_pretty /== nil and depth_printed fi_> 0 then
		cucharout(`(`);
		cucharout(`)`)
	else
		Pr_symbol(nil)
	endif
enddefine;


define lconstant Pr_procedure(pdr);
	lvars pdr, name, fni;
	if isarray(pdr) then
		fast_chain(pdr, Pr_array)
	endif;
	if isproperty(pdr) then
		fast_chain(pdr, Pr_hash_table)
	endif;
	if isbytespecpdr(pdr) then
		fast_chain(pdr, Pr_bytespec)
	endif;
	f_name(pdr) -> name;
	unless issymbol(name) or isstring(name) do
		fast_chain(pdr, Pr_pop11_simple)
	endunless;
	CHECK_PRINT_READABLY pdr;
	Pr_unreadable(
		popstackmark,
		if isundef(pdr) then
			'UNDEFINED'
		endif,
		if generic_function_p(pdr) then
			'GENERIC'
		endif,
		'FUNCTION',
		Pr_unreadable_quote, name,
		if print_escape /== nil
		and not(isundef(pdr)) then
			function_info(pdr) -> fni;
			lisp_true(f_min(fni));
			lisp_true(f_max(fni));
			lisp_true(f_results(fni))
		endif)
enddefine;


define lconstant Pr_process(ps);
	lvars ps, pid;
	CHECK_PRINT_READABLY ps;
	Pr_unreadable(
		popstackmark,
		if not(isliveprocess(ps)) then 'DEAD' endif,
		'PROCESS',
		if (process_pid(ps) ->> pid) then pid endif)
enddefine;


define lconstant Pr_readtable(r);
	lvars r;
	CHECK_PRINT_READABLY r;
	Pr_unreadable(
		popstackmark, readtable_case_sym(r), 'READTABLE', readtable_id(r))
enddefine;


define lconstant Pr_stream(stream);
	lvars stream, In, Out, name;
	CHECK_PRINT_READABLY stream;

	define lconstant Trim_trailing_nulls(s) -> s;
		lvars s;
		if isstring(s) then
			substring(1, skipchar_back(0, fast_vector_length(s), s) or 0, s)
				-> s
		endif
	enddefine;

	stream_source(stream) -> In;
	stream_dest(stream) -> Out;
	if isident(In) then idval(In) -> In endif;
	if isident(Out) then idval(Out) -> Out endif;
	if In == pop_undef then stream_read_p(stream) -> In endif;
	if Out == pop_undef then stream_write_p(stream) -> Out endif;

	Pr_unreadable(
		popstackmark,
		if isdevice(In) and (device_full_name(In) ->> name) then
			if In == Out then
				'FILE-IO-STREAM'
			else
				'FILE-INPUT-STREAM'
			endif;
			pathname(name)
		elseif isdevice(Out) and (device_full_name(Out) ->> name) then
			'FILE-OUTPUT-STREAM', pathname(name)
		elseif stringp(In) then
			'STRING-INPUT-STREAM', Pr_unreadable_quote, Trim_trailing_nulls(In)
		elseif stringp(Out) then
			'STRING-OUTPUT-STREAM', Pr_unreadable_quote, Trim_trailing_nulls(Out)
		elseif issymbol(In) then
			'SYNONYM-STREAM-TO', In
		elseif listp(In) then
			'CONCATENATED-STREAM', In
		elseif listp(Out) then
			'BROADCAST-STREAM', Out
		elseif In and Out then
			/* Note cannot distinguish echo streams */
			'TWO-WAY-STREAM', In, Out
		elseif In then
			'INPUT-STREAM', In
		elseif Out then
			'OUTPUT-STREAM', Out
		else
			'NULL-STREAM'
		endif)
enddefine;


define lconstant Pr_undef(u);
	lvars u;
	CHECK_PRINT_READABLY u;
	Pr_unreadable(popstackmark, 'UNDEF', if (undefword(u) ->> u) then u endif)
enddefine;


define lconstant Pr_termin() with_nargs 1;
	cucharout()
enddefine;


/* Set up key/print-procedure table */

define Pr_pop11_compound(item);
	lvars item, type;
	class_name(sys_class_of(item)) -> type;
	cucharout(`#`);
	cucharout(`S`);
	_lisppr(conspair(type, datalist(item)))
enddefine;


define lconstant Choose_print_procedure(key, prop);
	lvars key, prop;
	if is_vector_key(key) then
		Pr_vector ->> prop(key)
	elseif class_print(key) == sys_syspr then
		if is_record_key(key)
		and (key_of_dataword(class_dataword(key)) == key) then
			Pr_pop11_compound
		else
			Pr_pop11_simple
		endif
	else
		syspr
	endif
enddefine;


unless isproperty(lisp_class_print) do
	newproperty([], 32, false, true) -> lisp_class_print;
	Choose_print_procedure -> property_active(lisp_class_print)
endunless;


Pr_bitvector    -> lisp_class_print(bitvector_key);
Pr_integer      -> lisp_class_print(biginteger_key);
Pr_boolean      -> lisp_class_print(boolean_key);
Pr_character    -> lisp_class_print(character_key);
Pr_complex      -> lisp_class_print(complex_key);
Pr_float        -> lisp_class_print(ddecimal_key);
Pr_float        -> lisp_class_print(decimal_key);
Pr_integer      -> lisp_class_print(integer_key);
Pr_list         -> lisp_class_print(pair_key);
Pr_nil          -> lisp_class_print(nil_key);
Pr_package      -> lisp_class_print(package_key);
Pr_pathname     -> lisp_class_print(pathname_key);
Pr_procedure    -> lisp_class_print(procedure_key);
Pr_process      -> lisp_class_print(process_key);
Pr_random_state -> lisp_class_print(random_state_key);
Pr_ratio        -> lisp_class_print(ratio_key);
Pr_readtable    -> lisp_class_print(readtable_key);
Pr_stream       -> lisp_class_print(stream_key);
Pr_string       -> lisp_class_print(string_key);
Pr_symbol       -> lisp_class_print(symbol_key);
Pr_termin       -> lisp_class_print(termin_key);
Pr_undef        -> lisp_class_print(undef_key);
Pr_vector       -> lisp_class_print(vector_key);
sys_syspr       -> lisp_class_print(word_key);


endsection;


/* --- Revision History ---------------------------------------------------
--- John Williams, May 16 1995
		Added print_right_margin, print_lines, print_miser_width, and
		print_pprint_dispatch. (Last three not used yet).
--- John Williams, Apr 12 1995
		___________Pr_pathname modified to cope with logical pathnames.
--- John Williams, Apr  3 1995
		Changes for readtable case and print_readably.
--- John Williams, Mar 30 1995
		New version of _________Pr_stream.
--- John Williams, Dec 16 1994
		pr_sym_name now recognises print_case = nil (as set by the
		~(...~) format directive).
--- John Williams, Aug 25 1994
		Changes for Steele 1990 (Lisp version 1.6)
--- John Williams, Jun  7 1994
		_________Pr_symbol now copes with deleted packages.
--- John Williams, Apr 26 1994
		lispsyspr now uses PRINT-OBJECT.
--- John Williams, Feb 14 1994
		________Pr_array now prints a zero-dimensional array as #0A NIL, not #0ANIL.
--- John Williams, Dec 21 1993
		Now uses is_vector_key and is_record_key.
--- John Williams, Aug 11 1993
		___________Pr_pathname uses #p syntax. Added writeable where appropriate.
--- John Williams, Jan 16 1991
		Changed -class_spec- to -class_field_spec-
 */
