/* --- Copyright University of Sussex 1995. All rights reserved. ----------
 > File:            C.all/lisp/src/seq.p
 > Purpose:         Common Lisp sequence functions
 > Author:          John Williams, Dec 7 1986 (see revisions)
 > Documentation:   CLtL, p245-261
 > Related Files:   C.all/lisp/src/seq.lsp
 */

lisp_compile_mode;

section $-lisp;

constant procedure (copy_list, listp, nth);


/* Recognising sequences */

define vectorp(x);
	lvars x;
	if isvectorclass(x) then true else isarray(x) and arrayrank(x) == 1 endif
enddefine;


define stringp(x);
	lvars x;
	isstring(x)
		or
	(isarray(x) and arrayrank(x) == 1 and isstring(arrayvector(x)))
enddefine;


define bitvectorp(x);
	lvars x;
	isbitvector(x)
		or
	(isarray(x) and arrayrank(x) == 1 and isbitvector(arrayvector(x)))
enddefine;


define sequencep(x);
	lvars x;
	listp(x) or vectorp(x)
enddefine;


define is_array_seq(x) -> x;
	lvars x;
	unless (isarray(x) ->> x) and arrayrank(x) == 1 do
		false -> x
	endunless
enddefine;


define lconstant Seq_needed(item);
	lvars item;
	if item == [] then
		lisp_error('Non-empty sequence needed', [[]])
	else
		type_error(item, @SEQUENCE)
	endif
enddefine;


/* Checking sequence bounds */

define arrayseq_bounds(a) -> hi -> lo;
	lvars a, hi, lo, fp;
	arrayvector_bounds(a) -> lo -> hi;
	if isinteger(fast_fill_pointer(a) ->> fp) then
		fp fi_+ lo fi_- 1 -> hi
	endif
enddefine;


define Vseq_bounds(seq) -> (seq, lo, hi);
	lvars seq, lo, hi;
	if isvectorclass(seq) then
		1 -> lo;
		fast_vector_length(seq) -> hi
	elseif is_array_seq(seq) then
		arrayseq_bounds(seq) -> hi -> lo;
		arrayvector(seq) -> seq
	else
		Seq_needed(seq)
	endif
enddefine;


define check_seq_bounds(user_lo, user_hi, lo, hi) -> lo -> hi;
	lvars user_lo, user_hi, lo, hi, real_hi;
	if pop_true(user_hi) then
		hi -> real_hi;
		lo fi_+ check_positive(user_hi) fi_- 1 -> hi;
		if hi fi_> real_hi then
			lisp_error('End-index (~D) exceeds sequence length', [^user_hi])
		endif
	endif;
	if pop_true(user_lo) then
		lo fi_+ check_positive(user_lo) -> lo;
		if lo fi_> (hi fi_+ 1) then
			lisp_error('Start-index (~D) exceeds end-index (~D)',
					   [^user_lo ^user_hi])
		endif
	endif
enddefine;


define lconstant Check_hi_lo(hi, lo) -> n;
	lvars hi, lo, n;
	if pop_true(hi) then
		check_positive(hi) fi_- lo -> n;
		if n fi_< 0 then
			lisp_error('Start index (~D) exceeds end index (~D)', [^lo ^hi])
		endif
	else
		false -> n
	endif
enddefine;


define checkr_array_seq_index(i, seq) -> i;
	lvars i, seq, fp;
	if (fast_fill_pointer(seq) ->> fp)
	and i >= fp then
		lisp_error('Index (~D) exceeds fill-pointer (~D)', [^i ^fp])
	endif
enddefine;


/* Sequence access */

define elt(seq, i);
	lvars seq, i;
	if ispair(seq) then
		nth(i, seq)
	elseif isvectorclass(seq) then
		fast_vector_access(i, seq)
	elseif is_array_seq(seq) then
		fast_apply(checkr_array_seq_index(i, seq), seq)
	else
		Seq_needed(seq)
	endif
enddefine;


define updaterof elt(seq, i) with_nargs 3;
	lvars seq, i;
	if ispair(seq) then
		-> nth(i, seq)
	elseif isvectorclass(seq) then
		-> fast_vector_access(i, seq)
	elseif is_array_seq(seq) then
		-> fast_apply(checkr_array_seq_index(i, seq), seq)
	else
		Seq_needed(seq)
	endif
enddefine;


define seq_length(seq) -> len;
	lvars seq, len;
	if listp(seq) then
		0 -> len;
		until endp(seq) do
			len fi_+ 1 -> len;
			fast_back(seq) -> seq
		enduntil
	elseif isvectorclass(seq) then
		fast_vector_length(seq) -> len
	elseif is_array_seq(seq) then
		fast_fill_pointer(seq) or array_used_size(seq) -> len
	else
		Seq_needed(seq)
	endif
enddefine;


/* Copying portions of sequences */

define lconstant Seq_back(seq, i, flag);
	lvars seq, i, flag;
	if ispair(seq) then
		fast_back(seq)
	else
		lisp_error('~A index (~D) exceeds sequence length', [^flag ^i])
	endif
enddefine;


define lconstant Seq_destpair(seq, i, flag);
	lvars seq, i, flag;
	if ispair(seq) then
		fast_destpair(seq)
	else
		lisp_error('~A index (~D) exceeds sequence length', [^flag ^i])
	endif
enddefine;


define lconstant Subvseq(seq, start, finish) -> newseq;
	lvars seq, start, finish, newseq, lo, hi;
	Vseq_bounds(seq) -> (seq, lo, hi);
	check_seq_bounds(start, finish, lo, hi) -> start -> finish;
	finish fi_- start fi_+ 1 -> finish;
	class_init(fast_datakey(seq))(finish) -> newseq;
	move_subvector(start, seq, 1, newseq, finish)
enddefine;


define subseq(seq, lo, hi);
	lvars seq, lo, hi;
	if listp(seq) then
		fast_repeat check_positive(lo) times
			Seq_back(seq, lo, 'Start') -> seq
		endfast_repeat;
		if (Check_hi_lo(hi, lo) ->> lo) then
			[% fast_repeat lo times
				Seq_destpair(seq, hi, 'End') -> seq
			endfast_repeat %]
		else
			copy_list(seq)
		endif
	else
		Subvseq(seq, lo, hi)
	endif
enddefine;


define copy_seq(seq);
	lvars seq;
	if isvectorclass(seq) then
		copy(seq)
	elseif listp(seq) then
		copy_list(seq)
	else
		Subvseq(seq, 0, false)
	endif
enddefine;


define nreverse(seq) -> seq;
	lvars seq, i, j, procedure subscr, vec;
	if listp(seq) then
		seq -> i;
		[] -> seq;
		until endp(i) do
			(fast_back(i), i, seq) -> (i, seq, fast_back(i))
		enduntil
	else
		for Vseq_bounds(seq) -> (vec, i, j);
			class_fast_subscr(fast_datakey(vec)) -> subscr;
		step i fi_+ 1 -> i;
			 j fi_- 1 -> j
		till j fi_<= i do
			subscr(j, vec), subscr(i, vec) -> subscr(j, vec) -> subscr(i, vec)
		endfor
	endif
enddefine;


define reverse(seq) -> new;
	lvars seq, new;
	if listp(seq) then
		[] -> new;
		until endp(seq) do
			conspair(fast_destpair(seq) -> seq, new) -> new
		enduntil
	else
		nreverse(copy_seq(seq)) -> new
	endif
enddefine;


/* Constructing and exploding sequences */

define lconstant Checkr_stype_key(type) -> (key, xtype);
	lvars type, key, xtype;
	stype_->_key(type) -> (key, xtype);
	unless key do
		lisp_error('Invalid sequence type specifier', [^type])
	endunless
enddefine;


define lconstant Check_stype_size(xtype, size);
	lvars xtype, size, dims;
	if ispair(xtype) then
		nth(2, xtype) -> dims;
		if ispair(dims) then
			unless fast_back(dims) == []
			and (fast_front(dims) == @* or fast_front(dims) == size) do
				lisp_error(
					'Sequence length (~D) incompatible with type specifier',
					[^size ^xtype])
			endunless
		endif
	endif
enddefine;


define make_sequence(type, size, init);
	lvars type, size, init, key;
	Checkr_stype_key(type) -> (key, type);
	Check_stype_size(type, size);
	if init == nil then
		key_->_init(key) -> init
	endif;
	if key == string_key then
		char_code(init) -> init
	endif;
	if key == pair_key then
		check_positive(size) ->;
		conslist(fast_repeat size times init endfast_repeat, size)
	else
		initvectorclass(size, init, key)
	endif
enddefine;


define explode_seq(seq, for_string);
	lvars seq, for_string, temp;
	if ispair(seq) then
		if for_string then
			applist(seq, char_code)
		else
			destlist(seq) ->
		endif
	elseif isstring(seq) then
		if for_string then
			deststring(seq) ->
		else
			appdata(seq, fast_code_char)
		endif
	elseif isvectorclass(seq) then
		if for_string then
			appdata(seq, char_code)
		else
			explode(seq)
		endif
	elseif (is_array_seq(seq) ->> temp) then
		if for_string then
			appdata(temp, char_code)
		else
			explode(temp)
		endif;
		if isinteger(fast_fill_pointer(seq) ->> for_string) then
			erasenum(datalength(temp) fi_- for_string)
		endif
	elseunless seq == [] do
		Seq_needed(seq)
	endif
enddefine;


define dest_seq(seq, for_string);
	lvars seq, for_string;
	#| explode_seq(seq, for_string) |#
enddefine;


define fill_vseq(seq, for_string);
	lvars seq, for_string, lo, hi, key, procedure usubscr, i;
	Vseq_bounds(seq) -> (seq, lo, hi);
	fast_datakey(seq) -> key;
	if key == string_key and not(for_string) then
		#_< updater(lisp_subscrs) >_#
	else
		updater(class_fast_subscr(key))
	endif -> usubscr;
	fast_for i from hi by -1 to lo do
		usubscr(i, seq)
	endfast_for
enddefine;


define concatenate(type, seqs);
	lvars type, seqs, key, cons_pdr, for_string, len;
	Checkr_stype_key(type) -> (key, type);
	if key == pair_key then
		conslist
	else
		class_cons(key)
	endif -> cons_pdr;
	cons_pdr == consstring -> for_string;
	(#| until seqs == nil do
			explode_seq(fast_destpair(seqs) -> seqs, for_string)
		enduntil
	|#) -> len;
	Check_stype_size(type, len);
	fast_apply(len, cons_pdr)
enddefine;


/* Removing & deleting from sequences */




define lconstant Apply_seq_test(key, test) with_nargs 3;
	lvars key, test;
	lisp_apply((), key, 1, 1);
	lisp_apply(test, 1, 1);
	pop_true()
enddefine;


define lconstant List_remove_if(test, key, list, lo, hi, count, from_end) -> list;
	lvars test, key, list, lo, hi, count, from_end,
			start_list, removed, sl, n, item;

	list -> start_list;
	false -> removed;
	stacklength() -> sl;

	if pop_true(lo) then
		fast_repeat check_positive(lo) times
			Seq_destpair(list, lo, 'Start') -> list
		endfast_repeat
	else
		0 -> lo
	endif;

	if from_end then
		if (Check_hi_lo(hi, lo) ->> n) then
			fast_repeat n times
				Seq_destpair(list, hi, 'End') -> list
			endfast_repeat
		else
			destlist(list) -> n
		endif;
		fast_repeat n times
			quitif(count == 0);
			-> item;
			if Apply_seq_test(item, key, test) then
				unless removed do
					start_list -> list;
					fast_repeat stacklength() fi_- sl fi_+ 1 times
						fast_back(list) -> list
					endfast_repeat
				endunless;
				count fi_- 1 ->> count -> removed
			elseif removed then
				conspair(item, list) -> list
			endif
		endfast_repeat
	elseif (Check_hi_lo(hi, lo) ->> n) then
		fast_repeat n times
			quitif(count == 0);
			Seq_destpair(list, hi, 'End') -> list -> item;
			if Apply_seq_test(item, key, test) then
				count fi_- 1 ->> count -> removed
			else
				item
			endif
		endfast_repeat
	else
		until endp(list) or count == 0 do
			fast_destpair(list) -> list -> item;
			if Apply_seq_test(item, key, test) then
				count fi_- 1 ->> count -> removed
			else
				item
			endif
		enduntil
	endif;

	if removed then
		fast_repeat stacklength() fi_- sl times
			conspair(list) -> list
		endfast_repeat
	else
		setstacklength(sl);
		start_list -> list
	endif
enddefine;


define lconstant List_delete_if(test, key, list, lo, hi, count, from_end) -> result;
	lvars test, key, list, lo, hi, count, from_end, result, n, next;

	if pop_true(lo)
	and check_positive(lo) fi_> 0 then
		list -> result;
		fast_repeat lo fi_- 1 times
			Seq_back(list, lo, 'Start') -> list
		endfast_repeat
	else
		conspair(false, list) ->> list -> result;
		0 -> lo
	endif;

	if from_end then
		if (Check_hi_lo(hi, lo) ->> n) then
			fast_repeat (n fi_+ 1 ->> n) times
				list;
				Seq_back(list, hi, 'End') -> list
			endfast_repeat
		else
			0 -> n;
			until endp(list) do
				list;
				fast_back(list) -> list;
				n fi_+ 1 -> n
			enduntil
		endif;
		until n == 0 or count == 0 do
			-> list;
			if Apply_seq_test(fast_front(list), key, test) then
				count fi_- 1 -> count;
				->> next;
				fast_back(list) -> fast_back(next)
			endif;
			n fi_- 1 -> n
		enduntil;
		erasenum(n)

	elseif (Check_hi_lo(hi, lo) ->> n) then
		until n == 0 or count == 0 do
			n fi_- 1 -> n;
			Seq_back(list, hi, 'End') -> next;
			if Apply_seq_test(fast_front(next), key, test) then
				count fi_- 1 -> count;
				fast_back(next) -> fast_back(list)
			else
				next -> list
			endif
		enduntil
	else
		until count == 0 or endp(fast_back(list) ->> next) do
			if Apply_seq_test(fast_front(next), key, test) then
				count fi_- 1 -> count;
				fast_back(next) -> fast_back(list)
			else
				next -> list
			endif
		enduntil
	endif;

	if lo == 0 then
		fast_back(result) -> result
	endif
enddefine;


define lconstant Vector_remove_if(test, key, seq, lo, hi, count, from_end) -> seq;
	lvars test, key, seq, lo, hi, count, from_end,
			start, len, dkey, for_string, procedure subscr,
			removed, sl, i, item, templist;

	Vseq_bounds(seq) -> (seq, start, len);
	check_seq_bounds(lo, hi, start, len) -> lo -> hi;

	fast_datakey(seq) -> dkey;
	dkey == string_key -> for_string;
	class_fast_subscr(dkey) -> subscr;

	false -> removed;
	stacklength() -> sl;

	fast_for i from start to lo fi_- 1 do
		subscr(i, seq)
	endfast_for;
	if for_string then
		lisp_subscrs -> subscr
	endif;
	if from_end then
		[] -> templist;
		fast_for i from hi by -1 to lo do
			subscr(i, seq) -> item;
			if count fi_> 0 and Apply_seq_test(item, key, test) then
				count fi_- 1 -> count;
				true -> removed
			elseif for_string then
				conspair(char_code(item), templist) -> templist
			else
				conspair(item, templist) -> templist
			endif
		endfast_for;
		hi fi_+ 1 -> i
	else
		fast_for i from i to hi do
			quitif(count == 0);
			subscr(i, seq) -> item;
			if Apply_seq_test(item, key, test) then
				count fi_- 1 -> count;
				true -> removed
			elseif for_string then
				char_code(item)
			else
				item
			endif
		endfast_for
	endif;
	if removed then
		if from_end then
			while ispair(templist) do
				sys_grbg_destpair(templist) -> templist
			endwhile
		endif;
		if for_string then
			fast_subscrs -> subscr
		endif;
		fast_for i from i to len do
			subscr(i, seq)
		endfast_for;
		stacklength() fi_- sl -> sl;
		fast_apply(sl, class_cons(dkey)) -> seq
	else
		setstacklength(sl)
	endif
enddefine;


define remove_if(test, seq, from_end, lo, hi, count, key, ncopy);
	lvars test, seq, from_end, lo, hi, count, key, ncopy;

	checkr_function(test) -> test;
	checkr_function(key) -> key;
	;;; Assumes _____count is an integer
	pop_true(from_end) -> from_end;

	fast_apply(
		test, key, seq, lo, hi, count, from_end,
		if listp(seq) then
			if pop_true(ncopy) then
				List_delete_if
			else
				List_remove_if
			endif
		else
			Vector_remove_if
		endif)
enddefine;


/* Removing duplicates */

define lconstant Apply_seq_test2(x, y, key, test);
	lvars x, y, key, test;
	lisp_apply(x, key, 1, 1);
	lisp_apply(y, key, 1, 1);
	lisp_apply(test, 2, 1);
	pop_true()
enddefine;


define lconstant Isduplicate(item, list, key, test);
	lvars item, list, key, test;
	until endp(list) do
		if Apply_seq_test2(item, fast_destpair(list) -> list, key, test) then
			return(true)
		endif
	enduntil;
	false
enddefine;


define lconstant Isduplicate_from_end(item, list, key, test, here);
	lvars item, list, key, test, here;
	until list == here or endp(list) do
		if Apply_seq_test2(item, fast_destpair(list) -> list, key, test) then
			return(true)
		endif
	enduntil;
	false
enddefine;


define lconstant Del_dup(list, key, test) -> list;
	lvars list, key, test, next;
	until endp(fast_back(list) ->> next) do
		if Isduplicate(fast_front(next), fast_back(next), key, test) then
			fast_back(next) -> fast_back(list)
		else
			next -> list
		endif
	enduntil
enddefine;


define lconstant Del_dup_from_end(list, key, test) -> list;
	lvars list, key, test, start, next;
	fast_back(list) -> start;
	until endp(fast_back(list) ->> next) do
		if Isduplicate_from_end(fast_front(next), start, key, test, next) then
			fast_back(next) -> fast_back(list)
		else
			next -> list
		endif
	enduntil
enddefine;


define lconstant List_remove_duplicates(list, from_end, test, lo, hi, key);
	lvars list, from_end, test, lo, hi, key, result, work, lastwk, lastwk;

	conspair(false, nil) ->> result -> work;
	if pop_true(lo) then
		fast_repeat check_positive(lo) times
			conspair(Seq_destpair(list, lo, 'Start') -> list, nil)
				->> fast_back(work) -> work
		endfast_repeat
	endif;

	work -> lastwk;
	if (Check_hi_lo(hi, lo or 0) ->> lo) then
		fast_repeat lo times
			conspair(Seq_destpair(list, hi, 'End') -> list, nil)
				->> fast_back(lastwk) -> lastwk
		endfast_repeat
	else
		until endp(list) do
			conspair(fast_destpair(list) -> list, nil)
				->> fast_back(lastwk) -> lastwk
		enduntil
	endif;

	list -> fast_back(if pop_true(from_end) then
						  Del_dup_from_end(work, key, test)
					  else
						  Del_dup(work, key, test)
					  endif);
	fast_back(result)
enddefine;


define lconstant List_delete_duplicates(list, from_end, test, lo, hi, key) -> result;
	lvars list, from_end, test, lo, hi, key, result, work, n;

	if pop_true(lo) and check_positive(lo) fi_> 0 then
		list -> result;
		fast_repeat lo fi_- 1 times
			Seq_back(list, lo, 'Start') -> list
		endfast_repeat
	else
		0 -> lo;
		conspair(false, list) ->> result -> list
	endif;

	list -> work;
	if (Check_hi_lo(hi, lo) ->> n) then
		fast_repeat n times
			Seq_back(list, hi, 'End') -> list
		endfast_repeat;
		fast_back(list);
		nil -> fast_back(list)
	else
		nil
	endif -> fast_back(if pop_true(from_end) then
						   Del_dup_from_end(work, key, test)
					   else
						   Del_dup(work, key, test)
					   endif);

	if lo == 0 then
		fast_back(result) -> result
	endif
enddefine;


define remove_duplicates(seq, lo, hi, from_end, test, key, ncopy) -> seq;
	lvars seq, lo, hi, from_end, test, key, ncopy,
			start, len, dkey, procedure subscr, sl, i, list, newlen;

	checkr_function(key) -> key;
	checkr_function(test) -> test;

	if listp(seq) then
		if pop_true(ncopy) then
			List_delete_duplicates
		else
			List_remove_duplicates
		endif(seq, from_end, test, lo, hi, key) -> seq;
		return
	endif;

	Vseq_bounds(seq) -> (seq, start, len);
	check_seq_bounds(lo, hi, start, len) -> lo -> hi;
	fast_datakey(seq) -> dkey;
	class_fast_subscr(dkey) -> subscr;
	stacklength() -> sl;

	fast_for i from start to (lo fi_- 1) do
		subscr(i, seq)
	endfast_for;

	[% false,
	if dkey == string_key then
		fast_for i from i to hi do lisp_subscrs(i, seq) endfast_for
	else
		fast_for i from i to hi do subscr(i, seq) endfast_for
	endif %] -> list;

	if pop_true(from_end) then
		Del_dup_from_end
	else
		Del_dup
	endif (list, key, test) ->;

	dest_seq(fast_back(list), dkey == string_key) -> newlen;

	if pop_true(ncopy) and newlen == (hi fi_- lo fi_+ 1) then
		/* Nothing deleted - so return original */
		setstacklength(sl)
	else
		fast_for i from i to len do subscr(i, seq) endfast_for;
		class_cons(dkey)(stacklength() fi_- sl) -> seq
	endif
enddefine;


/* Sorting sequences */

constant procedure (symbol_function);

defprop
	numeric_sort_pred,
	char_sort_pred,
	stable_numeric_sort_pred,
	stable_char_sort_pred,
	;


define setup_sort_preds();
	/* This procedure is run after building the Lisp system, in
		C.all/lisp/src/clisp.p
	*/
	nonop <     -> numeric_sort_pred(symbol_function(@<));
	nonop <=    -> numeric_sort_pred(symbol_function(@<=));
	nonop >     -> numeric_sort_pred(symbol_function(@>));
	nonop >=    -> numeric_sort_pred(symbol_function(@>=));
	nonop <=    -> stable_numeric_sort_pred(symbol_function(@<=));
	nonop >=    -> stable_numeric_sort_pred(symbol_function(@>=));
	nonop fi_<  -> char_sort_pred(symbol_function(@CHAR<));
	nonop fi_<= -> char_sort_pred(symbol_function(@CHAR<=));
	nonop fi_>  -> char_sort_pred(symbol_function(@CHAR>));
	nonop fi_>= -> char_sort_pred(symbol_function(@CHAR>=));
	nonop fi_<= -> stable_char_sort_pred(symbol_function(@CHAR<=));
	nonop fi_>= -> stable_char_sort_pred(symbol_function(@CHAR>=));
enddefine;


define lconstant Stable_apply_seq_test2(x, y, key, test);
	lvars x, y, key, test;
	lisp_apply(x, key, 1, 1) -> x;
	lisp_apply(y, key, 1, 1) -> y;
	if pop_true(lisp_apply(x, y, test, 2, 1)) then
		true
	elseif pop_true(lisp_apply(y, x, test, 2, 1)) then
		false
	else
		1
	endif
enddefine;


define sort_seq(seq, pred, key, stable) -> seq;
	lvars seq, pred, key, stable, before_p, for_string = false, list;

	lconstant Before_p_clos
		= writeable Apply_seq_test2(% identfn, identfn %);

	dlocal
		% pdpart(Before_p_clos) %,
		% fast_frozval(1, Before_p_clos) %,
		% fast_frozval(2, Before_p_clos) %,
		;

	returnif(seq == []);
	checkr_function(pred) -> pred;
	if key == nil then
		;;; defaults
		identfn
	else
		checkr_function(key)
	endif -> key;

	if pop_true(stable) then
		unless key == identfn
		and (stable_numeric_sort_pred(pred) ->> before_p)
		or  ((stable_char_sort_pred(pred) ->> before_p)
				and
			 (stringp(seq) ->> for_string))
		do
			Stable_apply_seq_test2 -> pdpart(Before_p_clos);
			Before_p_clos -> before_p
		endunless
	else
		unless key == identfn
		and (numeric_sort_pred(pred) ->> before_p)
		or  ((char_sort_pred(pred) ->> before_p)
				and
			 (stringp(seq) ->> for_string))
		do
			Apply_seq_test2 -> pdpart(Before_p_clos);
			Before_p_clos -> before_p
		endunless
	endif;
	if before_p == Before_p_clos then
		key -> fast_frozval(1, Before_p_clos);
		pred -> fast_frozval(2, Before_p_clos)
	endif;

	if ispair(seq) then
		nc_listsort(seq, before_p) -> seq
	else
		/* Convert vector to a list, and sort that. Much quicker than
			using a "quicksort" algorithm on the vector directly
			(twice as quick for 5000 items).
		*/
		[% explode_seq(seq, for_string) %] -> list;
		nc_listsort(list, before_p) -> list;
		until list == [] do
			sys_grbg_destpair(list) -> list
		enduntil;
		fill_vseq(seq, for_string)
	endif
enddefine;


endsection;


/* --- Revision History ---------------------------------------------------
--- John Williams, Mar 15 1995
		Now signals typed errors.
--- John Williams, Aug 25 1994
		Changes for Steele 1990 (Lisp version 1.6)
--- John Williams, Sep  1 1993
		More improvements to sort_seq.
--- John Williams, Aug 27 1993
		Upgraded to Steele 1990.
		sort_seq now optimises common sort predicates.
--- John Williams, Jul 29 1993
		The ______stable option to sort_seq now works correctly.
		Updater of elt given correct pdnargs (3).
--- John Williams, Jul 28 1993
		Uses new nc_listsort.
--- John Williams, Jul 27 1993
		Fixed bug in updater of elt.
--- John Williams, Jul 12 1993
		No longer uses cons_with.
 */
