(* --- Copyright University of Sussex 1991. All rights reserved. ----------
 * File:            C.all/pml/lib/OS.ml
 * Purpose:         Operating System Interface
 * Author:          Robert John Duncan, Feb  1 1991
 * Documentation:   HELP * OS
 * Related Files:   C.all/pml/lib/OS.sig
 *)


external structure OS : OS = struct

uses sysdefs;

/* Exceptions */

ml_exception Error of string * string * string;
ml_exception NotImplemented of string;

lconstant
	procedure Error				= ml_valof("Error"),
	procedure NotImplemented	= ml_valof("NotImplemented"),
	procedure SOME				= ml_valof('Option.SOME'),
	NONE						= ml_valof('Option.NONE'),
;

define lconstant error(msg, culprits, name, arg);
	lvars msg, culprits, name, arg;
	if iscaller(sysfileok) then
		;;; assume a filename error
		'error in filename' -> msg;
	else
		sysiomessage() -> msg;
		if isstartstring('\s(', msg) then allbutfirst(2, msg) -> msg endif;
		if isendstring(')', msg) then allbutlast(1, msg) -> msg endif;
	endif;
	if not(arg) then nullstring -> arg endif;
	ml_raise(Error(ml_constuple(name, msg, arg, 3)));
enddefine;


/* O/S Version */

ml_val version : unit -> string =
procedure(unit) with_props version;
	lvars unit;
	allbutfirst(1, allbutlast(1, sys_os_type sys_>< nullstring));
endprocedure;

ml_val isunix : unit -> bool =
procedure(unit) with_props isunix;
	lvars unit;
	DEF UNIX;
endprocedure;

ml_val isvms : unit -> bool =
procedure(unit) with_props isvms;
	lvars unit;
	DEF VMS;
endprocedure;


/* The Current Process State */

ml_val arglist0: unit -> string list =
procedure(unit) with_props arglist0;
	lvars unit;
	maplist(poparglist0, copy);
endprocedure;

ml_val arglist : unit -> string list =
procedure(unit) with_props arglist;
	lvars unit;
	maplist(poparglist, copy);
endprocedure;

ml_val username : unit -> string =
procedure(unit) with_props username;
	lvars unit;
	popusername;
endprocedure;

ml_val translate : string -> string Option.option =
procedure(name)  with_props translate;
	lvars	name, value;
	dlocal	prmishap = error(% 'translate', name %);
	if systranslate(name) ->> value then
		SOME(value);
	else
		NONE;
	endif;
endprocedure;

ml_val pid : unit -> int =
procedure(unit) with_props pid;
	lvars unit;
	poppid;
endprocedure;

ml_val pwd : unit -> string =
procedure(unit) with_props pwd;
	lvars	unit;
	dlocal	prmishap = error(% 'pwd', false %);
	current_directory;
endprocedure;

ml_val cd : string -> unit =
procedure(dir) with_props cd;
	lvars	dir;
	dlocal	prmishap = error(% 'cd', dir %);
	dir -> current_directory;
	ml_unit;
endprocedure;

ml_val date : unit -> string =
procedure(unit) with_props date;
	lvars	unit;
	dlocal	prmishap = error(% 'date', false %);
	sysdaytime();
endprocedure;


/* Process Control */

ml_val obey : string -> int =
procedure(cmd) with_props obey;
	lvars	cmd;
	dlocal	prmishap = error(% 'obey', cmd %);
#_IF DEF UNIX
	if systranslate('SHELL') then cmd, `!` -> cmd endif;
#_ENDIF
	sysobey(cmd);
	pop_status;
endprocedure;

ml_val wait : unit -> (int * int) Option.option =
procedure(unit) with_props wait;
	lvars	unit, pid;
	dlocal	prmishap = error(% 'wait', false %);
	if syswait() ->> pid then
		SOME(ml_constuple(pid, pop_status, 2));
	else
		NONE;
	endif;
endprocedure;

ml_val kill : int -> bool =
procedure(pid) with_props kill;
	lvars	pid;
	dlocal	prmishap = error(% 'kill', false %);
	syskill(pid);
endprocedure;


/* UNIX-only */

ml_val fork : unit -> int Option.option =
procedure(unit) with_props fork;
	lvars	unit, pid;
#_IF DEF UNIX
	dlocal	prmishap = error(% 'fork', false %);
	if sysfork() ->> pid then
		SOME(pid);
	else
		NONE;
	endif;
#_ELSE
	ml_raise(NotImplemented('fork'));
#_ENDIF
endprocedure;

ml_val execve : string -> string list -> string list -> 'a =
procedure(cmd, arglist, envlist) with_props execve;
	lvars cmd, arglist, envlist;
#_IF DEF UNIX
	dlocal	prmishap = error(% 'execve', cmd %);
	sysexecute(cmd, arglist, envlist);
#_ELSE
	ml_raise(NotImplemented('execve'));
#_ENDIF
endprocedure;

ml_val envlist : unit -> string list =
procedure(unit) with_props envlist;
	lvars unit;
#_IF DEF UNIX
	popenvlist;
#_ELSE
	ml_raise(NotImplemented('envlist'));
#_ENDIF
endprocedure;


/* VMS-only */

ml_val spawn : string -> string -> string -> bool -> int Option.option =
procedure(cmd, input, output, wait) with_props spawn;
	lvars	cmd, input, output, wait, pid;
#_IF DEF VMS
	dlocal	prmishap = error(% 'spawn', cmd %);
	if cmd = nullstring then false -> cmd endif;
	if input = nullstring then false -> input endif;
	if output = nullstring then false -> output endif;
	if sysspawn(cmd, input, output, wait) ->> pid then
		SOME(pid);
	else
		NONE;
	endif;
#_ELSE
	ml_raise(NotImplemented('spawn'));
#_ENDIF
endprocedure;

ml_val attach : int -> bool =
procedure(pid) with_props attach;
	lvars	pid;
#_IF DEF VMS
	dlocal	prmishap = error(% 'attach', false %);
	sysattach(pid);
#_ELSE
	ml_raise(NotImplemented('attach'));
#_ENDIF
endprocedure;

end;	(* structure OS *)
